#include <string>
#include <memory>

using namespace std;

#include <cxcparam/parameter.h>

#include <Exception/Exception.h>

#include <cxcparamx/simple.h>

/**
   \class cxcparamx::simple
   \headerfile cxcparamx/simple.h

   A simple wrapper class providing polymorphic attribute retrieval.

   Instantiating the class opens the parameter file; the file is
   closed during object destruction.

**/

namespace cxcparamx {

    
    using pParamFileImpl = simple::pParamFileImpl;
    using ParamFileImpl = simple::ParamFileImpl ;

    static pParamFileImpl
    _init_parfile( const string& filename,
                   int argc, char** argv, const string& mode ) {

        const char* c_filename = filename.size() ? filename.c_str() : nullptr;

        // paramopen returns a ParamFile value, which is typedefed to
        // a pointer. unique_ptr needs to have a type that we can
        // point to, so had cxcparam defined two types (one as the
        // struct, one as the struct pointer) we could use that.  But
        // since it doesn't, we declare (but don't define!) struct
        // ParamFileImpl, and cast paramopen's ParamFile return value
        // to ParamFileImpl*.  As long as we only use a pointer, C++
        // doesn't need to have the struct defined, and all is good.

        pParamFileImpl pf( static_cast<ParamFileImpl*>( paramopen( c_filename, argv, argc, mode.c_str() ) ), 
                           []( ParamFileImpl* pf) { paramclose( pf ); } );

        if ( pf == nullptr )
            throw( Exception( ( filename.size() == 0 ? argv[0] : filename )
                              + ": error opening parameter file" ) );

        return pf;
    }

    /** \brief Constructor with file name taken from argv[0]
     *
     * \param[in] argc the number of elements in \c argv
     * \param[in] argv typically, the command line arguments vector passed to \c main*
     * \param[in] mode mode with which to open the file
     */
    simple::simple( int argc,  char** argv, const string& mode ) : pf( _init_parfile( "", argc, argv, mode ) ), filename( paramgetpath( pf.get() ) ) { }

    /** \brief Constructor with explicit parameter file name
     *
     * \param[in] filename the parameter file name
     * \param[in] argc the number of elements in \c argv
     * \param argv typically, the command line arguments vector passed to \c main
     * \param[in] mode mode with which to open the file
     */
    simple::simple( const string& in_filename,
                    int argc, char** argv, const string& mode ) : pf( _init_parfile( in_filename, argc, argv, mode ) ), filename( in_filename ) {

        if ( ! filename.size() )
            filename = paramgetpath( pf.get() );
    }

    bool simple::access( const string& name ) const {
        return paccess( pf.get(), name.c_str() );
    }

    void
    simple::access_t( const string& name ) const {
	if ( ! access( name ) )
	    throw( Exception( filename + ": no such parameter: " + name ) );
    }

    /** \brief Retrieve a parameter value into a float
     *
     * \param[in] name parameter name
     * \param[out]  value returned value
     * \throws Exception if parameter is not found
     */
    void
    simple::get( const string& name, float& value ) const {
	access_t(  name );
	value = pgetf( pf.get(), name.c_str() );
    }

    /** \brief Retrieve a parameter value into a double
     *
     * \param[in] name parameter name
     * \param[out]  value returned value
     * \throws Exception if parameter is not found
     */
    void
    simple::get( const string& name, double& value ) const {
        access_t( name );
	value = pgetd( pf.get(), name.c_str() );
    }


    /** \brief Retrieve a parameter value into a short
     *
     * \param[in] name parameter name
     * \param[out]  value returned value
     * \throws Exception if parameter is not found
     */
    void
    simple::get( const string& name, short& value ) const {
	access_t( name );
	value = pgets( pf.get(), name.c_str() );
    }

    /** \brief Retrieve a parameter value into an int
     *
     * \param[in] name parameter name
     * \param[out]  value returned value
     * \throws Exception if parameter is not found
     */
    void
    simple::get( const string& name, int& value ) const {
	access_t( name );
	value = pgeti( pf.get(), name.c_str() );
    }

    /** \brief Retrieve a parameter value into a long
     *
     * \param[in] name parameter name
     * \param[out]  value returned value
     * \throws Exception if parameter is not found
     */
    void
    simple::get( const string& name, long& value ) const {
        access_t( name );

	// pgetl is in cxcparam/parameter.h, but isn't in the actual
	// library
	value = pgeti( pf.get(), name.c_str() );
    }


    /** \brief Retrieve a parameter value into an bool
     *
     * \param[in] name parameter name
     * \param[out]  value returned value
     * \throws Exception if parameter is not found
     */
    void
    simple::get( const string& name, bool& value ) const {
        access_t( name );
        value = pgetb( pf.get(), name.c_str() );
    }

    /** \brief Retrieve a parameter value into an string
     *
     * \param[in] name parameter name
     * \param[out]  value returned value
     * \throws Exception if parameter is not found
     */
    void
    simple::get( const string& name, string& value ) const {

	access_t( name );

	const int strlen = 1024;
	char buffer[strlen+1];

	pgetstr( pf.get(), name.c_str(), buffer, strlen );
	value = buffer;
    }


    //////////////////////////////////////////////

    /** \brief Return a parameter value as a float
     *
     * \param[in] name parameter name
     * \throws Exception if parameter is not found
     */
    float
    simple::getf( const string& name ) const {
        access_t( name );
	return pgetf( pf.get(), name.c_str() );
    }

    /** \brief Return a parameter value as a double
     *
     * \param[in] name parameter name
     * \throws Exception if parameter is not found
     */
    double
    simple::getd( const string& name ) const {
	access_t( name );
	return pgetd( pf.get(), name.c_str() );
    }


    /** \brief Return a parameter value as a short
     *
     * \param[in] name parameter name
     * \throws Exception if parameter is not found
     */
    short
    simple::gets( const string& name ) const {
	access_t( name );
	return pgets( pf.get(), name.c_str() );
    }

    /** \brief Return a parameter value as an int
     *
     * \param[in] name parameter name
     * \throws Exception if parameter is not found
     */
    int
    simple::geti( const string& name ) const {
	access_t( name );
	return pgeti( pf.get(), name.c_str() );
    }

    /** \brief Return a parameter value as a long
     *
     * \param[in] name parameter name
     * \throws Exception if parameter is not found
     */
    long
    simple::getl( const string& name ) const {
	access_t( name );

	// pgetl is in cxcparam/parameter.h, but isn't in the actual
	// library
	return pgeti( pf.get(), name.c_str() );
    }


    /** \brief Return a parameter value as an bool
     *
     * \param[in] name parameter name
     * \throws Exception if parameter is not found
     */
    bool
    simple::getb( const string& name ) const {
	access_t( name );
	return pgetb( pf.get(), name.c_str() );
    }

    /** \brief Return a parameter value as an string
     *
     * \param[in] name parameter name
     * \throws Exception if parameter is not found
     */
    string
    simple::getstr( const string& name ) const {

	access_t( name );

	const int strlen = 1024;
	char buffer[strlen+1];

	pgetstr( pf.get(), name.c_str(), buffer, strlen );
	return string(buffer);
    }
}
